#!/usr/bin/env python
from __future__ import division, print_function

import inspect
import sys

import astropy.io.fits as pyfits

if len(sys.argv) != 4:
	sys.stderr.write("""SYNOPSIS: %s <catalogue.fits> <tablename> <skyarea>

tablename: name of the catalogue
skyarea: catalogue area in square degrees

Author: Johannes Buchner (C) 2013-2025
""" % sys.argv[0])
	sys.exit(1)

f = pyfits.open(sys.argv[1])
print('current', f[1].name, 'SKYAREA:', f[1].header.get('SKYAREA', None))
assert '_' not in sys.argv[2], 'Table name must not contain underscore "_".'
f[1].name = sys.argv[2]
f[1].header['SKYAREA'] = float(sys.argv[3])
print('new    ', f[1].name, 'SKYAREA:', f[1].header.get('SKYAREA', None))

args = inspect.signature(pyfits.writeto).parameters if hasattr(inspect, 'signature') else inspect.getargspec(pyfits.writeto).args
if 'overwrite' in args:
	arg_overwrite = 'overwrite'
else:
	arg_overwrite = 'clobber'
f.writeto(sys.argv[1], **{arg_overwrite:True})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division, print_function

import sys
import warnings

from . import progress


class FakeProgressBar(object):
	def __init__(self, *args):
		pass

	def __call__(self, it):
		return it

	def start(self):
		return self

	def increment(self):
		pass

	def finish(self):
		pass


class NullOutputLogger(object):
	def __init__(self):
		pass

	def log(self, *msg):
		pass

	def warn(self, msg):
		warnings.warn(msg, stacklevel=3)

	def progress(self, *args, **kwargs):
		return FakeProgressBar()


class NormalLogger(object):
	def __init__(self):
		pass

	def log(self, msg):
		sys.stderr.write('%s\n' % msg)

	def warn(self, msg):
		warnings.warn(msg, stacklevel=3)

	def progress(self, ndigits=6, *args, **kwargs):
		return progress.bar(ndigits=ndigits, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Explain a multiway association of astrometric catalogue. Use --help for usage.

Example: nway-explain.py out.fits 179
"""

from __future__ import division, print_function

import argparse
import sys

import astropy.io.fits as pyfits
import matplotlib.pyplot as plt
import numpy
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.collections import PatchCollection
from matplotlib.patches import Ellipse


class HelpfulParser(argparse.ArgumentParser):
	def error(self, message):
		sys.stderr.write('error: %s\n' % message)
		self.print_help()
		sys.exit(2)


parser = HelpfulParser(description=__doc__,
	epilog="""Johannes Buchner (C) 2013-2025 <johannes.buchner.acad@gmx.com>""",
	formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('matchcatalogue', type=str,
	help="""nway output catalogue""")

parser.add_argument('id', type=str,
	help='ID to explain (from primary catalogue)')


# parsing arguments
args = parser.parse_args()

#print('loading catalogue %s' % args.matchcatalogue)
f = pyfits.open(args.matchcatalogue)
header = f[0].header
data = f[1].data
primary_id_col = header['COL_PRIM']
#print('    searching for %s == %s' % (primary_id_col, args.id))
if issubclass(data.dtype[primary_id_col].type, numpy.integer):
	mask = data[primary_id_col] == int(args.id)
elif issubclass(data.dtype[primary_id_col].type, float):
	mask = data[primary_id_col] == float(args.id)
else:
	mask = data[primary_id_col] == args.id
#print('    %d rows found' % (mask.sum()))
if mask.sum() == 0:
	print('ERROR: ID not found. Was searching for %s == %s' % (primary_id_col, args.id))
	sys.exit(1)
#print()
# make a plot of the positions

plt.figure(figsize=(12,12))
plt.axis('equal')
cols_ra = header['COLS_RA'].split(' ')
cols_dec = header['COLS_DEC'].split(' ')
cols_err = header['COLS_ERR'].split(' ')
tablenames = header['TABLES'].split(', ')
center_ra = data[cols_ra[0]][mask][0]
center_dec = data[cols_dec[0]][mask][0]
p_any = data['p_any'][mask][0]

print('NWAY results for Source %s:' % args.id)
print()
if p_any > 0.8:
	print('This source probably has a counterpart (p_any=%.2f)' % p_any)
elif p_any < 0.1:
	print('This source probably does not a counterpart (p_any=%.2f)' % p_any)
else:
	print('It is uncertain if this source has a counterpart (p_any=%.2f)' % p_any)
print()
print("Assuming it has a counterpart, we have the following possible associations:")
print()

j_option = 0


def print_option(name, i):
	global j_option
	j_option += 1
	matchflag = data['match_flag'][mask][i]
	matchflagstars = '**' if matchflag == 1 else ('*' if matchflag == 2 else '')
	if p_any < 0.1:
		matchflagstars = ''
	if matchflag == 0:
		print('Association %d: probability p_i=%.2f ' % (j_option, data['p_i'][mask][i]))
	else:
		print('Association %d%s[match_flag==%d]: probability p_i=%.2f ' % (j_option, matchflagstars, matchflag, data['p_i'][mask][i]))
	print('     Involved catalogues:  %s ' % (name))
	for col in header['BIASING'].split(', '):
		if col.strip() == '':
			continue
		bias = data['bias_' + col][mask][i]
		if bias >= 2:
			print('     prior %-15s increased the probability (bias_%s=%.2f)' % (col, col, bias))
		elif bias <= 0.5:
			print('     prior %-15s decreased the probability (bias_%s=%.2f)' % (col, col, bias))
	print()


def convx(ra):
	return (ra - center_ra) * 3600


def convy(dec):
	return (dec - center_dec) * 3600


def converr(err):
	return err * 3600


markers = ['x', '+', '^', '<', '>', 'v', 'p'] * 10
colors = ['b', 'c', 'g', 'r', 'k', 'brown'] * 10
all_options = []
ii = numpy.argsort(data['p_i'][mask])[::-1]
for col_ra, col_dec, col_err, marker, color in zip(cols_ra, cols_dec, cols_err, markers, colors):
	tblname, err = col_err.split('_', 1)
	if err.startswith(':'):
		err_components = err[1:].split(':')
		if len(err_components) == 1:
			ra_err = data[tblname + '_' + err_components[0]][mask]
			dec_err = ra_err
			pa_err = numpy.zeros(ra_err.shape)
		elif len(err_components) == 2:
			ra_err = data[tblname + '_' + err_components[0]][mask]
			dec_err = data[tblname + '_' + err_components[1]][mask]
			pa_err = numpy.zeros(ra_err.shape)
		elif len(err_components) == 3:
			ra_err = data[tblname + '_' + err_components[0]][mask]
			dec_err = data[tblname + '_' + err_components[1]][mask]
			pa_err = data[tblname + '_' + err_components[2]][mask]
		else:
			assert len(err_components) == 3, err_components
	else:
		ra_err = numpy.ones(mask.sum()) * float(err)
		dec_err = numpy.ones(mask.sum()) * float(err)
		pa_err = numpy.zeros(ra_err.shape)
	pos = set(zip(data[col_ra][mask], data[col_dec][mask], ra_err, dec_err, pa_err))
	ras = numpy.array([ra for ra, dec, ra_err, dec_err, pa_err in pos if ra != -99])
	decs = numpy.array([dec for ra, dec, ra_err, dec_err, pa_err in pos if ra != -99])
	ra_errs = numpy.array([ra_err for ra, dec, ra_err, dec_err, pa_err in pos if ra != -99]) / 60. / 60.
	dec_errs = numpy.array([dec_err for ra, dec, ra_err, dec_err, pa_err in pos if ra != -99]) / 60. / 60.
	pa_errs = numpy.array([pa_err for ra, dec, ra_err, dec_err, pa_err in pos if ra != -99])
	r,  = plt.plot(convx(ras), convy(decs), marker=marker, mec=color, mfc='None', ms=8, mew=2, ls=' ', label='%s %s' % (col_ra, col_dec))
	patches = [
		Ellipse((convx(ra), convy(dec)),
		2 * converr(ra_err),
		2 * converr(dec_err),
		angle=90 - pa_err)
		for ra, dec, ra_err, dec_err, pa_err in zip(ras, decs, ra_errs, dec_errs, pa_errs)]
	p = PatchCollection(patches)
	p.set_facecolor('None')
	p.set_edgecolor(color)
	plt.gca().add_collection(p)
	options = [(-99, -99)]
	for i in ii:
		ra, dec = data[col_ra][mask][i], data[col_dec][mask][i]
		if (ra, dec) not in options:
			options.append((ra, dec))
	all_options.append(options)


def graph_make(all_options, highlight=None):
	for j, options in enumerate(all_options):
		if j != 0:
			plt.plot(j, 0, marker='x', color='gray')
		for k in range(len(options)-1):
			plt.plot(j, -(k + 1), marker='o', color='k')


def graph_highlight(all_options, selected):
	x = numpy.arange(len(all_options))
	plt.plot(x, selected, '-', color='k')


out_options = []
outfilename = '%s_explain_%s_options.pdf' % (args.matchcatalogue, args.id)
pp = PdfPages(outfilename)
maxy = max([len(o)-1 for o in all_options])
maxx = len(all_options)-1
for i in ii:
	plt.figure(figsize=(3+maxx, 3))
	plt.axis('off')
	graph_make(all_options)
	j = []
	name = []
	for col_ra, col_dec, options, tablename in zip(cols_ra, cols_dec, all_options, tablenames):
		radec = data[col_ra][mask][i], data[col_dec][mask][i]
		plt.text(len(j), 0.1, col_ra + '\n' + col_dec,
			rotation=90, size=6, ha='center', va='bottom')
		k = options.index(radec)
		j.append(-k)
		if k == 0:
			name.append("")
		elif k == 1:
			name.append("%s" % tablename)
	print_option('-'.join(name), i)
	graph_highlight(all_options, j)
	plt.text(-0.1, -1, 'p_i=%.2f' % data['p_i'][mask][i], ha='right', va='center')
	plt.text(maxx + 0.1, 0, r'$\leftarrow$ absent', ha='left', va='center')
	plt.ylim(-maxy-0.5, 0.5)
	plt.xlim(-0.5, maxx+0.5)
	plt.savefig(pp, format='pdf', bbox_inches='tight')
	plt.close()
pp.close()
print('plotting to %s' % outfilename)

# go through each association and highlight
for j, i in enumerate(numpy.argsort(data['p_single'][mask])[::-1][:3]):
	ras = []
	decs = []
	for col_ra, col_dec, marker in zip(cols_ra, cols_dec, markers):
		if data[col_ra][mask][i] == -99:
			continue
		ra = data[col_ra][mask][i]
		dec = data[col_dec][mask][i]
		ras.append(ra)
		decs.append(dec)

	plt.plot(convx(ras), convy(decs), '-', lw=(3-j), label='top %s by distance (p_single=%.2f, %d cat.)' % (j+1, data['p_single'][mask][i], data['ncat'][mask][i]), color='y')

mask2 = numpy.logical_and(mask, data['match_flag'] == 1)
ras = []
decs = []
first = True
for col_ra, col_dec, marker in zip(cols_ra, cols_dec, markers):
	ra = float(data[col_ra][mask2])
	dec = float(data[col_dec][mask2])
	if ra == -99:
		continue
	if not first:
		plt.text(convx(ra), convy(dec), ' 1', va='top', ha='left', alpha=0.5, size=16, fontweight='bold')
	first = False
	if ra == -99:
		continue
	ras.append(ra)
	decs.append(dec)
plt.plot(convx(ras), convy(decs), '-', lw=1.7, label='p_i=%.2f (match_flag=1)' % (float(data['p_i'][mask2])), color='orange')

mask2 = numpy.logical_and(mask, data['match_flag'] == 2)
for i in numpy.where(mask2)[0]:
	ras = []
	decs = []
	first = True
	for col_ra, col_dec, marker in zip(cols_ra, cols_dec, markers):
		ra = data[col_ra][i]
		dec = data[col_dec][i]
		if ra == -99:
			continue
		if not first:
			plt.text(convx(ra), convy(dec),
				' 2', va='top', ha='left', alpha=0.5,
				size=16, fontweight='bold')
		first = False
		ras.append(ra)
		decs.append(dec)
	plt.plot(convx(ras), convy(decs), '-', lw=0.5, label='p_i=%.2f (match_flag=2)' % (data['p_i'][i]), color='yellow')

plt.xlabel(r'$\Delta$RA [arcsec]')
plt.ylabel(r'$\Delta$DEC [arcsec]')
plt.title('Source %s, p_any=%.2f' % (args.id, p_any))
xlo, xhi = plt.xlim()
ylo, yhi = plt.ylim()
hi = max(-xlo, xhi, -ylo, yhi)
plt.ylim(-hi, hi) # DEC
plt.xlim(hi, -hi) # RA goes the other way
plt.legend(loc='best', numpoints=1, prop=dict(size=8))
outfilename = '%s_explain_%s.pdf' % (args.matchcatalogue, args.id)
print('plotting to %s' % outfilename)
print()
print("Disclaimer: These results assume that the input (sky densities, positional errors, and priors) are correct.")
print()
plt.savefig(outfilename, bbox_inches='tight')
plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division, print_function

__doc__ = """Create a fake, random-position catalogue for testing the false association rate.

For each source, tries to find a new position, by choosing randomly one of its K
nearest neighbors and picking a random location between them. If the new
location is within --radius (arcsec) of an old or new source, the random process
is repeated. With 2/3 probability K=10, with 1/3 probability K=100, giving a
good balance between reproducing local structures and filling the field.

Example: nway-create-fake-catalogue.py --radius 20 COSMOS-XMM.fits fake-COSMOS-XMM.fits
"""

import argparse
import sys

import astropy.io.fits as pyfits
import healpy
import matplotlib.pyplot as plt
import numpy
from numpy import (arccos, arcsin, arctan, arctan2, cos, exp, log10,
                   logical_and, pi, sin, sqrt, tan)

import nwaylib.fastskymatch as match
import nwaylib.progress as progress


class HelpfulParser(argparse.ArgumentParser):
	def error(self, message):
		sys.stderr.write('error: %s\n' % message)
		self.print_help()
		sys.exit(2)


parser = HelpfulParser(description=__doc__,
	epilog="""Johannes Buchner (C) 2013-2025 <johannes.buchner.acad@gmx.com>""",
	formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('--radius', type=float, required=True,
	help='Remove sources which are near original sources, within this radius (arcsec).')

parser.add_argument('inputfile', type=str, help="input catalogue fits file")
parser.add_argument('outputfile', help='output catalogue fits file')

# parsing arguments
args = parser.parse_args()

outfile = args.outputfile
filename = args.inputfile
radius = args.radius

print('opening', filename)
inputfitsfile = pyfits.open(filename)
header_hdu = inputfitsfile[0]
table = inputfitsfile[1].data

ra_key = match.get_tablekeys(table, 'RA')
print('    using RA  column: %s' % ra_key)
dec_key = match.get_tablekeys(table, 'DEC')
print('    using DEC column: %s' % dec_key)

ra_orig = table[ra_key]
dec_orig = table[dec_key]
ra = ra_orig + 0
dec = dec_orig + 0
n = len(ra_orig)

i_select = numpy.random.randint(0, n, size=400)
ra_test = ra[i_select]
dec_test = dec[i_select]
phi_test = ra_test / 180 * pi
theta_test = dec_test / 180 * pi + pi/2.
phi = ra / 180 * pi
theta = dec / 180 * pi + pi/2.

# find a pixelation that has ~20 in each pixel
print('finding good pixelation...')
ntarget = 20

nside = 1
for nside_next in range(30):
	# number of pixels
	nside = 2**nside_next
	npix = healpy.pixelfunc.nside2npix(nside)
	if n > ntarget * npix:
		# too few pixels, can not achieve target number per pixel
		continue
	i = healpy.pixelfunc.ang2pix(nside, phi=phi_test, theta=theta_test, nest=True)

	k = healpy.pixelfunc.ang2pix(nside, phi=phi, theta=theta, nest=True)

	nneighbors = []
	for rai, deci, ii in zip(ra_test, dec_test, i):
		nneighbors.append((k == ii).sum())

	nneighbors_total = sum(nneighbors)

	print('  nside=%d: the %d test objects have a total of %d neighbors' % (nside, len(ra_test), nneighbors_total))
	if nneighbors_total < len(ra_test) * ntarget:
		# small enough, accept
		print('    accepting.')
		break


def greatarc_interpolate(posa, posb, f):
	(a_ra, a_dec), (b_ra, b_dec) = posa, posb
	lon1 = a_ra / 180 * pi
	lat1 = a_dec / 180 * pi
	lon2 = b_ra / 180 * pi
	lat2 = b_dec / 180 * pi

	d = arccos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(lon1 - lon2))
	A = sin((1 - f) * d) / sin(d)
	B = sin(f * d) / sin(d)
	x = A * cos(lat1) * cos(lon1) + B * cos(lat2) * cos(lon2)
	y = A * cos(lat1) * sin(lon1) + B * cos(lat2) * sin(lon2)
	z = A * sin(lat1) + B * sin(lat2)

	lat_f = arctan2(z, sqrt(x**2 + y**2))
	lon_f = arctan2(y, x)

	c_ra = lon_f * 180 / pi
	c_dec = lat_f * 180 / pi
	return c_ra, c_dec


# for each of them, create a new one without collision
pbar = progress.bar(ndigits=6)
for index in pbar(range(n)):
	# select random coordinate
	#a = numpy.random.randint(0, n)
	a = index
	# select randomly from 10 nearest neighbors
	# find neighbors
	i = healpy.pixelfunc.ang2pix(nside, phi=phi[a], theta=theta[a], nest=True)
	j = healpy.pixelfunc.get_all_neighbours(nside, phi=phi[a], theta=theta[a], nest=True)
	neighbors = numpy.hstack((i, j))
	#print(neighbors.shape, k.shape, neighbors.shape)
	is_neighbor = (k.reshape((-1,1)) == neighbors.reshape((1,-1))).any(axis=1)
	#print('found %d neighbors' % is_neighbor.sum())
	ra_nearby = ra_orig[is_neighbor]
	dec_nearby = dec_orig[is_neighbor]

	d = match.dist((ra_orig[a], dec_orig[a]), (ra_nearby, dec_nearby))
	b_nearest = numpy.argsort(d)
	dmask = d[b_nearest] * 60 * 60 > radius
	#d = d[dmask]
	b_nearest = b_nearest[dmask]

	assert len(b_nearest) > 0, "Method failed: No sources found nearby, could not interpolate a fake source."

	#print('have %d neighbors' % len(b_nearest))
	b_nearest = b_nearest[:100]
	while True:
		if numpy.random.randint(0,3) == 0:
			b = b_nearest[numpy.random.randint(0, len(b_nearest))]
		else:
			b = b_nearest[numpy.random.randint(0, min(10, len(b_nearest)))]

		# compute point in between
		di = d[b]
		uexclude = radius / 60 / 60 / di
		u = numpy.random.uniform(uexclude, 1 - uexclude)
		ra_i, dec_i = greatarc_interpolate((ra_orig[a], dec_orig[a]), (ra_nearby[b], dec_nearby[b]), u)

		# check for collision with original catalogue
		d = match.dist((ra_i, dec_i), (ra_nearby, dec_nearby))
		if (d * 60 * 60 < radius).any():
			# print('rejecting, near a original source')
			continue  # try again
		# check for collision with new sources?
		d = match.dist((ra_i, dec_i), (ra[:i], dec[:i]))
		if (d * 60 * 60 < radius).any():
			# print('rejecting, near a new source')
			continue  # try again
		ra[index] = numpy.fmod(ra_i + 360, 360)
		dec[index] = dec_i
		break

table[ra_key] = ra
table[dec_key] = dec

tbhdu = inputfitsfile[1]
print('writing "%s" (%d rows)' % (outfile, len(tbhdu.data)))

hdulist = pyfits.HDUList([header_hdu, tbhdu])
hdulist.writeto(outfile, **progress.kwargs_overwrite_true)



"""
Probabilistic Cross-Identification of Astronomical Sources

Reference: Budavari & Szalay (2008), ApJ, 679:301-309
Authors: Johannes Buchner (C) 2013-2020
Authors: Tamas Budavari (C) 2012
"""
from __future__ import division, print_function

import numpy
from numpy import e, log, log10, pi

# use base 10 everywhere

log_arcsec2rad = log(3600 * 180 / pi)


def log_posterior(prior, log_bf):
	"""
	Returns log10 posterior probability normalised against alternative hypothesis
	that the sources are unrelated (Budavari+08)
	"""
	return -log10(1 + (1 - prior) * 10**(-log_bf - log10(prior)))


def posterior(prior, log_bf):
	"""
	Returns posterior probability normalised against alternative hypothesis
	that the sources are unrelated (Budavari+08)
	"""
	with numpy.errstate(over='ignore'):
		return 1. / (1 + (1 - prior) * 10**(-log_bf - log10(prior)))


def unnormalised_log_posterior(prior, log_bf, ncat):
	"""
	Returns posterior probability (without normalisation against alternatives)
	"""
	return log_bf + log10(prior)


def log_bf2(psi, s1, s2):
	"""
	log10 of the 2-way Bayes factor, see eq.(16)
	psi separation
	s1 and s2=accuracy of coordinates
	"""
	s = s1 * s1 + s2 * s2
	return (log(2) + 2 * log_arcsec2rad - log(s) - psi * psi / 2 / s) * log10(e)


def log_bf3(p12,p23,p31, s1,s2,s3):
	"""
	log10 of the 3-way Bayes factor, see eq.(17)
	"""
	ss1 = s1*s1
	ss2 = s2*s2
	ss3 = s3*s3
	s = ss1*ss2 + ss2*ss3 + ss3*ss1
	q = ss3 * p12**2 + ss1 * p23**2 + ss2 * p31**2
	return (log(4) + 4 * log_arcsec2rad - log(s) - q / 2 / s) * log10(e)


def log_bf(p, s):
	"""
	log10 of the multi-way Bayes factor, see eq.(18)

	p: separations matrix (NxN matrix of arrays)
	s: errors (list of N arrays)
	"""

	n = len(s)
	# precision parameter w = 1/sigma^2
	w = [numpy.asarray(si, dtype=float)**-2. for si in s]
	norm = (n - 1) * log(2) + 2 * (n - 1) * log_arcsec2rad
	del s

	wsum = numpy.sum(w, axis=0)
	slog = numpy.sum(log(w), axis=0) - log(wsum)
	q = 0
	for i, wi in enumerate(w):
		for j, wj in enumerate(w):
			if i < j:
				q += wi * wj * p[i][j]**2
	exponent = - q / 2 / wsum
	return (norm + slog + exponent) * log10(e)

# vectorized in the following means that many 2D matrices/vectors are going to be handled.
# i.e., each entry in the matrix or vector, is a vector of numbers.


def assert_possemdef(M):
	"""Check that the 2x2 matrix M is positive semi-definite, vectorized"""
	tr = M[0][0] + M[1][1]
	det = M[0][0] * M[1][1] - M[0][1] * M[0][1]
	mask = numpy.isclose(tr**2, 4 * det)
	if mask.all():
		return
	# bad square root:
	mask_badsqrt = numpy.logical_and(~mask, tr**2 < 4 * det)
	assert not numpy.any(mask_badsqrt), (tr[mask_badsqrt], det[mask_badsqrt], M)

	ev1 = (tr[~mask] + (tr[~mask]**2 - 4 * det[~mask])**0.5) / 2
	ev2 = (tr[~mask] - (tr[~mask]**2 - 4 * det[~mask])**0.5) / 2

	assert (ev1 >= 0).all(), ("EV1:", ev1[~(ev1 > 0)], tr[~mask][~(ev1 > 0)], det[~mask][~(ev1 > 0)], M)
	assert (ev2 >= 0).all(), ("EV2:", ev2[~(ev2 > 0)], tr[~mask][~(ev2 > 0)], det[~mask][~(ev2 > 0)], M)


def matrix_add(A, B):
	""" 2D matrix addition, vectorized """
	# assert_possemdef(A)
	# assert_possemdef(B)
	(a11, a12), (a21, a22) = A
	(b11, b12), (b21, b22) = B
	M = (a11 + b11, a12 + b12), (a21 + b21, a22 + b22)
	# assert_possemdef(M)
	return M


def matrix_multiply(A, B):
	""" 2D matrix multiplication, vectorized """
	(a11, a12), (a21, a22) = A
	(b11, b12), (b21, b22) = B
	M = (a11*b11+a12*b21, a11*b12+a12*b22), (a21*b11+a22*b21, a21*b12+a22*b22)
	return M


def matrix_invert(A):
	""" invert a matrix A, vectorized """
	(a, b), (c, d) = A
	F = 1.0 / (a * d - c * b)
	assert (F > 0).all()
	return (F * d, -F * b), (-F * c, F * a)


def matrix_det(A):
	""" 2D matrix determinant, vectorized """
	(a, b), (c, d) = A
	return a * d - b * c


def apply_vector_right(A, b):
	""" multiply 2D vector with 2D matrix, vectorized """
	(a11, a12), (a21, a22) = A
	(b1, b2) = b
	return a11*b1+a12*b2, a21*b1+a22*b2


def apply_vector_left(a, B):
	""" multiply 2D matrix with 2D vector, vectorized """
	a1, a2 = a
	(b11, b12), (b21, b22) = B
	return a1*b11+a2*b21, a1*b12+a2*b22


def vector_multiply(a, b):
	""" multiply two vectors, vectorized """
	a1, a2 = a
	b1, b2 = b
	return a1*b1+a2*b2


def vector_normalised(v):
	vnorm = (v[0]**2 + v[1]**2)**0.5
	return (
		numpy.where(vnorm == 0, 2**-0.5, v[0] / (vnorm + 1e-300)),
		numpy.where(vnorm == 0, 2**-0.5, v[1] / (vnorm + 1e-300)),
	)


def apply_vABv(v, A, B):
	""" compute v^T A B v, vectorized """
	return vector_multiply(v, apply_vector_right(matrix_add(A, B), v))
	#return vector_multiply(apply_vector_left(v, A), apply_vector_right(B, v))


def make_covmatrix(sigma_x, sigma_y, rho=0):
	""" create covariance matrix from given standard deviations and normalised correlation rho, vectorized """
	return (sigma_x**2, rho * sigma_x * sigma_y), (rho * sigma_x * sigma_y, sigma_y**2)


def make_invcovmatrix(sigma_x, sigma_y, rho=0):
	""" create inverse covariance matrix from given standard deviations and normalised correlation rho, vectorized """
	F = 1.0 / (sigma_x**2 * sigma_y**2 * (1 - rho**2))
	return (F * sigma_y**2, F * -rho * sigma_x * sigma_y), \
		(F * -rho * sigma_x * sigma_y, F * sigma_x**2)


def convert_from_ellipse(a, b, phi):
	""" create covariance parameters from ellipse major axis a, minor axis b and angle phi; vectorized

	as in Pineau+16, eq 8-10, for example.
	"""
	a2 = a**2
	b2 = b**2
	s = numpy.sin(phi)
	c = numpy.cos(phi)
	s2 = s**2
	c2 = c**2
	sigma_x = (a2 * s2 + b2 * c2)**0.5
	sigma_y = (a2 * c2 + b2 * s2)**0.5
	rho = c * s * (a2 - b2) / (sigma_x * sigma_y)
	return sigma_x, sigma_y, rho


def log_bf_elliptical(separations_ra, separations_dec, pos_errors):
	"""
	log10 of the multi-way Bayes factor, see eq.(18)

	separations_ra: RA separations matrix (NxN matrix of arrays)
	separations_dec: DEC separations matrix (NxN matrix of arrays)
	pos_errors: errors (list of (N,3) arrays) giving sigma_RA, sigma_DEC, rho
	"""

	error_matrices = [make_invcovmatrix(si, sj, rho)
		for si, sj, rho in pos_errors]

	circ_pos_errors = [((si**2 + sj**2) / 2)**0.5 for si, sj, rho in pos_errors]

	new_separations = [[None for j in range(len(error_matrices))] for i in range(len(error_matrices))]

	for i, Mi in enumerate(error_matrices):
		for j, Mj in enumerate(error_matrices):
			if i < j:
				v = (separations_ra[i][j], separations_dec[i][j])
				# get separation length
				d2 = vector_multiply(v, v)
				d = d2**0.5
				# get error in direction of separation
				vnormed = vector_normalised(v)
				wi = vector_multiply(apply_vector_left(vnormed, Mi), vnormed)
				wj = vector_multiply(apply_vector_left(vnormed, Mj), vnormed)
				# ratio of circular error to directional error:
				# combine the uncertainties by adding the variances
				dist_ratio = (circ_pos_errors[i]**2 + circ_pos_errors[j]**2) / (1/wi + 1/wj)
				# provide new separation
				new_separations[i][j] = d * dist_ratio**-0.5

	return log_bf(new_separations, circ_pos_errors)



"""
Computes adaptively binned histograms of two distributions.

Provides smooth biasing functions and plots them.

This is not specific for magnitudes, it can be any provided property.
"""

import matplotlib.pyplot as plt
import numpy
import scipy
import scipy.integrate
import scipy.interpolate
import scipy.signal
from numpy import exp, log10, logical_and, pi


def ratio(hist_sel, hist_all):
	"""
	ratio between the two histograms. Minor offsets to avoid zero and inf.
	"""
	with numpy.errstate(divide='ignore', invalid='ignore'):
		return numpy.where(hist_all == 0, 100, hist_sel / hist_all)


def fraction(bin_mag, hist_sel, hist_all):
	"""
	fraction of selected
	"""
	# need a re-normalisation so that unknown values can be set to prob=1
	m = hist_all > 0
	hist_sel_m = hist_sel[m]
	hist_all_m = hist_all[m]
	ratio = hist_sel_m / hist_all_m
	delta = (bin_mag[1:] - bin_mag[:-1])[m]
	avg = (ratio * hist_all_m * delta).sum() / (delta * hist_all_m).sum()
	assert avg != 0

	r = numpy.ones(len(hist_all))
	r[m] = hist_sel_m / hist_all_m / avg
	assert numpy.isfinite(r).all(), r[~numpy.isfinite(r)]
	assert (r > 0).all(), r[~(r > 0)]
	return r


def plot_fit(bin_mag, hist_sel, hist_all, func, name):
	"""
	Plotting
	"""
	mags = numpy.linspace(bin_mag.min(), bin_mag.max(), 400)
	plt.figure()
	plt.subplot(2, 1, 1)
	hist_n = ratio(hist_sel, hist_all)
	plt.plot(bin_mag[:-1], hist_all, '-',
		drawstyle='steps-post', label='all')
	plt.plot(bin_mag[:-1], hist_sel, '-',
		drawstyle='steps-post', label='selected')
	plt.legend(loc='best')
	plt.ylabel('normalized weight')
	plt.xlim(mags.min(), mags.max())
	plt.subplot(2, 1, 2)
	plt.plot(bin_mag[:-1], hist_n, '-',
		drawstyle='steps-post', label='ratio')
	plt.plot(mags, func(mags), '-', label='fit')
	plt.legend(loc='best')
	plt.ylabel('normalized weight')
	plt.xlabel(name)
	plt.xlim(mags.min(), mags.max())
	plt.yscale('log')
	plt.savefig(name.replace(':', '_') + '_fit.pdf', bbox_inches='tight')
	plt.close()


def fitfunc_histogram(bin_mag, hist_sel, hist_all):
	"""
	creates the biasing functions
	"""
	bin_n = ratio(hist_sel, hist_all)
	# w = scipy.signal.gaussian(5, 1)
	# w /= w.sum()
	# bin_n_smooth = scipy.signal.convolve(bin_n, w, mode='same')
	# no smoothing
	bin_n_smooth = bin_n
	interpfunc = scipy.interpolate.interp1d(bin_mag,
		list(bin_n_smooth) + [bin_n_smooth[-1]],
		bounds_error=False, kind='zero')
	return interpfunc


def adaptive_histograms(mag_all, mag_sel, weights=None):
	"""
	creates the histograms for the two columns in an adaptive way (based on mag_sel)
	with the same binning.
	"""
	if weights is None:
		weights = numpy.ones(len(mag_sel))
	assert len(weights) == len(mag_sel), (len(weights), len(mag_sel))
	mag_sel_idx = numpy.argsort(mag_sel)
	mag_sel_sorted = mag_sel[mag_sel_idx]
	weight_axis = numpy.cumsum(weights[mag_sel_idx]) / numpy.sum(weights)
	weight_axis[0] = 0
	weight_axis[-1] = 1
	# weight_axis = numpy.linspace(0, 1, len(mag_sel_sorted))

	func_sel = scipy.interpolate.interp1d(weight_axis, mag_sel_sorted)
	# choose bin borders based on cumulative distribution, using 15 points
	x = numpy.unique(func_sel(numpy.linspace(0, 1, 15)))
	lo, hi = numpy.nanmin(mag_all), numpy.nanmax(mag_all)
	# extend to make sure ratios are well-defined
	if x[-1] < hi:
		x = numpy.asarray(list(x) + [hi+1])
	if x[0] > lo:
		x = numpy.asarray([lo-1] + list(x))
	# linear histogram (for no adaptiveness):
	# x = numpy.linspace(mag_all.min(), mag_all.max(), 20)
	hist_sel, bins = numpy.histogram(mag_sel, bins=x,    density=True, weights=weights)
	hist_all, bins = numpy.histogram(mag_all, bins=bins, density=True)
	return bins, hist_sel, hist_all



import os

try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

long_description = ""
with open('README.rst') as f:
	long_description = f.read()

setup(
	packages=['nwaylib'],
	scripts=['nway.py', 'nway-write-header.py', 'nway-explain.py', 'nway-create-fake-catalogue.py', 'nway-create-shifted-catalogue.py', 'nway-calibrate-cutoff.py'],
	long_description=long_description,
)
